@@ -13,45 +13,94 @@
 ##############################################################################
 """Schema Fields
 """
-
 __docformat__ = 'restructuredtext'
 
-import re
+from datetime import datetime
+from datetime import date
+from datetime import timedelta
+from datetime import time
 import decimal
-from datetime import datetime, date, timedelta, time
+import re
+import threading
+
 from zope.event import notify
+from zope.interface import classImplements
+from zope.interface import implementer
+from zope.interface import Interface
+from zope.interface.interfaces import IInterface
+from zope.interface.interfaces import IMethod
 
-from zope.interface import classImplements, implements, Interface
-from zope.interface.interfaces import IInterface, IMethod
-
-from zope.schema.interfaces import IField
-from zope.schema.interfaces import IMinMaxLen, IText, ITextLine
-from zope.schema.interfaces import ISourceText
-from zope.schema.interfaces import IInterfaceField
-from zope.schema.interfaces import IBytes, IASCII, IBytesLine, IASCIILine
-from zope.schema.interfaces import IBool, IInt, IFloat, IDatetime, IFrozenSet
-from zope.schema.interfaces import IChoice, ITuple, IList, ISet, IDict
-from zope.schema.interfaces import IPassword, IDate, ITimedelta
-from zope.schema.interfaces import IObject, IBeforeObjectAssignedEvent
-from zope.schema.interfaces import ITime, IDecimal
-from zope.schema.interfaces import IURI, IId, IDottedName, IFromUnicode
-from zope.schema.interfaces import ISource, IBaseVocabulary
+from zope.schema.interfaces import IASCII
+from zope.schema.interfaces import IASCIILine
+from zope.schema.interfaces import IBaseVocabulary
+from zope.schema.interfaces import IBeforeObjectAssignedEvent
+from zope.schema.interfaces import IBool
+from zope.schema.interfaces import IBytes
+from zope.schema.interfaces import IBytesLine
+from zope.schema.interfaces import IChoice
 from zope.schema.interfaces import IContextSourceBinder
+from zope.schema.interfaces import IDate
+from zope.schema.interfaces import IDatetime
+from zope.schema.interfaces import IDecimal
+from zope.schema.interfaces import IDict
+from zope.schema.interfaces import IDottedName
+from zope.schema.interfaces import IField
+from zope.schema.interfaces import IFloat
+from zope.schema.interfaces import IFromUnicode
+from zope.schema.interfaces import IFrozenSet
+from zope.schema.interfaces import IId
+from zope.schema.interfaces import IInt
+from zope.schema.interfaces import IInterfaceField
+from zope.schema.interfaces import IList
+from zope.schema.interfaces import IMinMaxLen
+from zope.schema.interfaces import IObject
+from zope.schema.interfaces import IPassword
+from zope.schema.interfaces import ISet
+from zope.schema.interfaces import ISource
+from zope.schema.interfaces import ISourceText
+from zope.schema.interfaces import IText
+from zope.schema.interfaces import ITextLine
+from zope.schema.interfaces import ITime
+from zope.schema.interfaces import ITimedelta
+from zope.schema.interfaces import ITuple
+from zope.schema.interfaces import IURI
 
-from zope.schema.interfaces import ValidationError, InvalidValue
-from zope.schema.interfaces import WrongType, WrongContainedType, NotUnique
-from zope.schema.interfaces import SchemaNotProvided, SchemaNotFullyImplemented
-from zope.schema.interfaces import InvalidURI, InvalidId, InvalidDottedName
+from zope.schema.interfaces import ValidationError
+from zope.schema.interfaces import InvalidValue
+from zope.schema.interfaces import WrongType
+from zope.schema.interfaces import WrongContainedType
+from zope.schema.interfaces import NotUnique
+from zope.schema.interfaces import SchemaNotProvided
+from zope.schema.interfaces import SchemaNotFullyImplemented
+from zope.schema.interfaces import InvalidURI
+from zope.schema.interfaces import InvalidId
+from zope.schema.interfaces import InvalidDottedName
 from zope.schema.interfaces import ConstraintNotSatisfied
 
-from zope.schema._bootstrapfields import Field, Container, Iterable, Orderable
-from zope.schema._bootstrapfields import Text, TextLine, Bool, Int, Password
+from zope.schema._bootstrapfields import Field
+from zope.schema._bootstrapfields import Container  # API import for __init__
+from zope.schema._bootstrapfields import Iterable
+from zope.schema._bootstrapfields import Orderable
+from zope.schema._bootstrapfields import Text
+from zope.schema._bootstrapfields import TextLine
+from zope.schema._bootstrapfields import Bool
+from zope.schema._bootstrapfields import Int
+from zope.schema._bootstrapfields import Password
 from zope.schema._bootstrapfields import MinMaxLen
 from zope.schema.fieldproperty import FieldProperty
 from zope.schema.vocabulary import getVocabularyRegistry
 from zope.schema.vocabulary import VocabularyRegistryError
 from zope.schema.vocabulary import SimpleVocabulary
 
+from zope.schema._compat import b
+from zope.schema._compat import text_type
+from zope.schema._compat import string_types
+from zope.schema._compat import binary_type
+from zope.schema._compat import PY3
+from zope.schema._compat import make_binary
+
+# pep 8 friendlyness
+Container
 
 # Fix up bootstrap field types
 Field.title = FieldProperty(IField['title'])
@@ -72,59 +121,37 @@ classImplements(Bool, IFromUnicode)
 classImplements(Int, IInt)
 
 
+@implementer(ISourceText)
 class SourceText(Text):
     __doc__ = ISourceText.__doc__
-    implements(ISourceText)
-    _type = unicode
+    _type = text_type
 
 
+@implementer(IBytes, IFromUnicode)
 class Bytes(MinMaxLen, Field):
     __doc__ = IBytes.__doc__
-    implements(IBytes, IFromUnicode)
 
-    _type = str
+    _type = binary_type
 
-    def fromUnicode(self, u):
+    def fromUnicode(self, uc):
+        """ See IFromUnicode.
         """
-        >>> b = Bytes(constraint=lambda v: 'x' in v)
-
-        >>> b.fromUnicode(u" foo x.y.z bat")
-        ' foo x.y.z bat'
-        >>> b.fromUnicode(u" foo y.z bat")
-        Traceback (most recent call last):
-        ...
-        ConstraintNotSatisfied:  foo y.z bat
-
-        """
-        v = str(u)
+        v = make_binary(uc)
         self.validate(v)
         return v
 
+# for things which are of the str type on both Python 2 and 3
+if PY3:  # pragma: no cover
+    NativeString = Text
+else:  # pragma: no cover
+    NativeString = Bytes
 
-class ASCII(Bytes):
+
+@implementer(IASCII)
+class ASCII(NativeString):
     __doc__ = IASCII.__doc__
-    implements(IASCII)
 
     def _validate(self, value):
-        """
-        >>> ascii = ASCII()
-
-        Make sure we accept empty strings:
-
-        >>> empty = ''
-        >>> ascii._validate(empty)
-
-        and all kinds of alphanumeric strings:
-
-        >>> alphanumeric = "Bob\'s my 23rd uncle"
-        >>> ascii._validate(alphanumeric)
-
-        >>> umlauts = "Köhlerstraße"
-        >>> ascii._validate(umlauts)
-        Traceback (most recent call last):
-        ...
-        InvalidValue
-        """
         super(ASCII, self)._validate(value)
         if not value:
             return
@@ -132,90 +159,77 @@ class ASCII(Bytes):
             raise InvalidValue
 
 
+@implementer(IBytesLine)
 class BytesLine(Bytes):
     """A Text field with no newlines."""
 
-    implements(IBytesLine)
-
     def constraint(self, value):
         # TODO: we should probably use a more general definition of newlines
-        return '\n' not in value
+        return b('\n') not in value
+
+# for things which are of the str type on both Python 2 and 3
+if PY3:  # pragma: no cover
+    NativeStringLine = TextLine
+else:  # pragma: no cover
+    NativeStringLine = BytesLine
 
 
+@implementer(IASCIILine)
 class ASCIILine(ASCII):
     __doc__ = IASCIILine.__doc__
 
-    implements(IASCIILine)
-
     def constraint(self, value):
         # TODO: we should probably use a more general definition of newlines
         return '\n' not in value
 
 
+@implementer(IFloat, IFromUnicode)
 class Float(Orderable, Field):
     __doc__ = IFloat.__doc__
-    implements(IFloat, IFromUnicode)
     _type = float
 
     def __init__(self, *args, **kw):
         super(Float, self).__init__(*args, **kw)
 
-    def fromUnicode(self, u):
+    def fromUnicode(self, uc):
+        """ See IFromUnicode.
         """
-        >>> f = Float()
-        >>> f.fromUnicode("1.25")
-        1.25
-        >>> f.fromUnicode("1.25.6") #doctest: +IGNORE_EXCEPTION_DETAIL
-        Traceback (most recent call last):
-        ...
-        ValueError: invalid literal for float(): 1.25.6
-        """
-        v = float(u)
+        v = float(uc)
         self.validate(v)
         return v
 
 
+@implementer(IDecimal, IFromUnicode)
 class Decimal(Orderable, Field):
     __doc__ = IDecimal.__doc__
-    implements(IDecimal, IFromUnicode)
     _type = decimal.Decimal
 
     def __init__(self, *args, **kw):
         super(Decimal, self).__init__(*args, **kw)
 
-    def fromUnicode(self, u):
-        """
-        >>> f = Decimal()
-        >>> import decimal
-        >>> isinstance(f.fromUnicode("1.25"), decimal.Decimal)
-        True
-        >>> float(f.fromUnicode("1.25"))
-        1.25
-        >>> f.fromUnicode("1.25.6")
-        Traceback (most recent call last):
-        ...
-        ValueError: invalid literal for Decimal(): 1.25.6
+    def fromUnicode(self, uc):
+        """ See IFromUnicode.
         """
         try:
-            v = decimal.Decimal(u)
+            v = decimal.Decimal(uc)
         except decimal.InvalidOperation:
-            raise ValueError('invalid literal for Decimal(): %s' % u)
+            raise ValueError('invalid literal for Decimal(): %s' % uc)
         self.validate(v)
         return v
 
 
+@implementer(IDatetime)
 class Datetime(Orderable, Field):
     __doc__ = IDatetime.__doc__
-    implements(IDatetime)
     _type = datetime
 
     def __init__(self, *args, **kw):
         super(Datetime, self).__init__(*args, **kw)
 
 
+@implementer(IDate)
 class Date(Orderable, Field):
     __doc__ = IDate.__doc__
-    implements(IDate)
     _type = date
 
     def _validate(self, value):
@@ -224,48 +238,56 @@ class Date(Orderable, Field):
             raise WrongType(value, self._type, self.__name__)
 
 
+@implementer(ITimedelta)
 class Timedelta(Orderable, Field):
     __doc__ = ITimedelta.__doc__
-    implements(ITimedelta)
     _type = timedelta
 
 
+@implementer(ITime)
 class Time(Orderable, Field):
     __doc__ = ITime.__doc__
-    implements(ITime)
     _type = time
 
 
+@implementer(IChoice, IFromUnicode)
 class Choice(Field):
     """Choice fields can have a value found in a constant or dynamic set of
     values given by the field definition.
     """
-    implements(IChoice, IFromUnicode)
 
     def __init__(self, values=None, vocabulary=None, source=None, **kw):
         """Initialize object."""
         if vocabulary is not None:
-            assert (isinstance(vocabulary, basestring)
-                    or IBaseVocabulary.providedBy(vocabulary))
-            assert source is None, (
-                "You cannot specify both source and vocabulary.")
+            if (not isinstance(vocabulary, string_types)
+                    and not IBaseVocabulary.providedBy(vocabulary)):
+                raise ValueError('vocabulary must be a string or implement '
+                                 'IBaseVocabulary')
+            if source is not None:
+                raise ValueError(
+                    "You cannot specify both source and vocabulary.")
         elif source is not None:
             vocabulary = source
 
-        assert not (values is None and vocabulary is None), (
-               "You must specify either values or vocabulary.")
-        assert values is None or vocabulary is None, (
-               "You cannot specify both values and vocabulary.")
+        if (values is None and vocabulary is None):
+            raise ValueError(
+                "You must specify either values or vocabulary."
+            )
+        if values is not None and vocabulary is not None:
+            raise ValueError(
+                "You cannot specify both values and vocabulary."
+            )
 
         self.vocabulary = None
         self.vocabularyName = None
         if values is not None:
             self.vocabulary = SimpleVocabulary.fromValues(values)
-        elif isinstance(vocabulary, (unicode, str)):
+        elif isinstance(vocabulary, string_types):
             self.vocabularyName = vocabulary
         else:
-            assert (ISource.providedBy(vocabulary) or
-                    IContextSourceBinder.providedBy(vocabulary))
+            if (not ISource.providedBy(vocabulary)
+                    and not IContextSourceBinder.providedBy(vocabulary)):
+                raise ValueError('Invalid vocabulary')
             self.vocabulary = vocabulary
         # Before a default value is checked, it is validated. However, a
         # named vocabulary is usually not complete when these fields are
@@ -285,27 +307,17 @@ class Choice(Field):
         # get registered vocabulary if needed:
         if IContextSourceBinder.providedBy(self.vocabulary):
             clone.vocabulary = self.vocabulary(object)
-            assert ISource.providedBy(clone.vocabulary)
         elif clone.vocabulary is None and self.vocabularyName is not None:
             vr = getVocabularyRegistry()
             clone.vocabulary = vr.get(object, self.vocabularyName)
-            assert ISource.providedBy(clone.vocabulary)
+
+        if not ISource.providedBy(clone.vocabulary):
+            raise ValueError('Invalid clone vocabulary')
 
         return clone
 
     def fromUnicode(self, str):
-        """
-        >>> from vocabulary import SimpleVocabulary
-        >>> t = Choice(
-        ...     vocabulary=SimpleVocabulary.fromValues([u'foo',u'bar']))
-        >>> IFromUnicode.providedBy(t)
-        True
-        >>> t.fromUnicode(u"baz")
-        Traceback (most recent call last):
-        ...
-        ConstraintNotSatisfied: baz
-        >>> t.fromUnicode(u"foo")
-        u'foo'
+        """ See IFromUnicode.
         """
         self.validate(str)
         return str
@@ -323,12 +335,112 @@ class Choice(Field):
             except VocabularyRegistryError:
                 raise ValueError("Can't validate value without vocabulary")
         if value not in vocabulary:
-            raise ConstraintNotSatisfied(value)
+            raise ConstraintNotSatisfied(value, self.__name__)
 
 
+_isuri = r"[a-zA-z0-9+.-]+:"  # scheme
+_isuri += r"\S*$"  # non space (should be pickier)
+_isuri = re.compile(_isuri).match
+
+
+@implementer(IURI, IFromUnicode)
+class URI(NativeStringLine):
+    """URI schema field
+    """
+
+    def _validate(self, value):
+        super(URI, self)._validate(value)
+        if _isuri(value):
+            return
+
+        raise InvalidURI(value)
+
+    def fromUnicode(self, value):
+        """ See IFromUnicode.
+        """
+        v = str(value.strip())
+        self.validate(v)
+        return v
+
+
+_isdotted = re.compile(
+    r"([a-zA-Z][a-zA-Z0-9_]*)"
+    r"([.][a-zA-Z][a-zA-Z0-9_]*)*"
+    # use the whole line
+    r"$").match
+
+
+@implementer(IDottedName)
+class DottedName(NativeStringLine):
+    """Dotted name field.
+
+    Values of DottedName fields must be Python-style dotted names.
+    """
+
+    def __init__(self, *args, **kw):
+        self.min_dots = int(kw.pop("min_dots", 0))
+        if self.min_dots < 0:
+            raise ValueError("min_dots cannot be less than zero")
+        self.max_dots = kw.pop("max_dots", None)
+        if self.max_dots is not None:
+            self.max_dots = int(self.max_dots)
+            if self.max_dots < self.min_dots:
+                raise ValueError("max_dots cannot be less than min_dots")
+        super(DottedName, self).__init__(*args, **kw)
+
+    def _validate(self, value):
+        """
+
+        """
+        super(DottedName, self)._validate(value)
+        if not _isdotted(value):
+            raise InvalidDottedName(value)
+        dots = value.count(".")
+        if dots < self.min_dots:
+            raise InvalidDottedName(
+                "too few dots; %d required" % self.min_dots, value
+            )
+        if self.max_dots is not None and dots > self.max_dots:
+            raise InvalidDottedName("too many dots; no more than %d allowed" %
+                                    self.max_dots, value)
+
+    def fromUnicode(self, value):
+        v = value.strip()
+        if not isinstance(v, self._type):
+            v = v.encode('ascii')
+        self.validate(v)
+        return v
+
+
+@implementer(IId, IFromUnicode)
+class Id(NativeStringLine):
+    """Id field
+
+    Values of id fields must be either uris or dotted names.
+    """
+
+    def _validate(self, value):
+        super(Id, self)._validate(value)
+        if _isuri(value):
+            return
+        if _isdotted(value) and "." in value:
+            return
+
+        raise InvalidId(value)
+
+    def fromUnicode(self, value):
+        """ See IFromUnicode.
+        """
+        v = value.strip()
+        if not isinstance(v, self._type):
+            v = v.encode('ascii')
+        self.validate(v)
+        return v
+
+
+@implementer(IInterfaceField)
 class InterfaceField(Field):
     __doc__ = IInterfaceField.__doc__
-    implements(IInterfaceField)
 
     def _validate(self, value):
         super(InterfaceField, self)._validate(value)
@@ -347,23 +459,25 @@ def _validate_sequence(value_type, value, errors=None):
 
     To illustrate, we'll use a text value type. All values must be unicode.
 
-            >>> field = TextLine(required=True)
+       >>> from zope.schema._compat import u
+       >>> from zope.schema._compat import b
+       >>> field = TextLine(required=True)
 
-        To validate a sequence of various values:
+    To validate a sequence of various values:
 
-            >>> errors = _validate_sequence(field, ('foo', u'bar', 1))
-            >>> errors
-            [WrongType('foo', <type 'unicode'>, ''), WrongType(1, <type 'unicode'>, '')]
+       >>> errors = _validate_sequence(field, (b('foo'), u('bar'), 1))
+       >>> errors # XXX assumes Python2 reprs
+       [WrongType('foo', <type 'unicode'>, ''), WrongType(1, <type 'unicode'>, '')]
 
-        The only valid value in the sequence is the second item. The others
-        generated errors.
+    The only valid value in the sequence is the second item. The others
+    generated errors.
 
-        We can use the optional errors argument to collect additional errors
-        for a new sequence:
+    We can use the optional errors argument to collect additional errors
+    for a new sequence:
 
-        >>> errors = _validate_sequence(field, (2, u'baz'), errors)
-        >>> errors
-        [WrongType('foo', <type 'unicode'>, ''), WrongType(1, <type 'unicode'>, ''), WrongType(2, <type 'unicode'>, '')]
+       >>> errors = _validate_sequence(field, (2, u('baz')), errors)
+       >>> errors # XXX assumes Python2 reprs
+       [WrongType('foo', <type 'unicode'>, ''), WrongType(1, <type 'unicode'>, ''), WrongType(2, <type 'unicode'>, '')]
 
     """
     if errors is None:
@@ -373,7 +487,7 @@ def _validate_sequence(value_type, value, errors=None):
     for item in value:
         try:
             value_type.validate(item)
-        except ValidationError, error:
+        except ValidationError as error:
             errors.append(error)
     return errors
 
@@ -417,41 +531,44 @@ class AbstractCollection(MinMaxLen, Iterable):
             _validate_uniqueness(value)
 
 
+@implementer(ITuple)
 class Tuple(AbstractCollection):
     """A field representing a Tuple."""
-    implements(ITuple)
     _type = tuple
 
 
+@implementer(IList)
 class List(AbstractCollection):
     """A field representing a List."""
-    implements(IList)
     _type = list
 
 
+@implementer(ISet)
 class Set(AbstractCollection):
     """A field representing a set."""
-    implements(ISet)
     _type = set
 
     def __init__(self, **kw):
-        if 'unique' in kw: # set members are always unique
+        if 'unique' in kw:  # set members are always unique
             raise TypeError(
                 "__init__() got an unexpected keyword argument 'unique'")
         super(Set, self).__init__(unique=True, **kw)
 
 
+@implementer(IFrozenSet)
 class FrozenSet(AbstractCollection):
-    implements(IFrozenSet)
     _type = frozenset
 
     def __init__(self, **kw):
-        if 'unique' in kw: # set members are always unique
+        if 'unique' in kw:  # set members are always unique
             raise TypeError(
                 "__init__() got an unexpected keyword argument 'unique'")
         super(FrozenSet, self).__init__(unique=True, **kw)
 
 
+VALIDATED_VALUES = threading.local()
+
+
 def _validate_fields(schema, value, errors=None):
     if errors is None:
         errors = []
@@ -462,11 +579,12 @@ def _validate_fields(schema, value, errors=None):
     if schema is Interface:
         return errors
     # if `value` is part of a cyclic graph, we need to break the cycle to avoid
-    # infinite recursion.
-    if hasattr(value, '__schema_being_validated'):
+    # infinite recursion. Collect validated objects in a thread local dict by
+    # it's python represenation. A previous version was setting a volatile
+    # attribute which didn't work with security proxy
+    if id(value) in VALIDATED_VALUES.__dict__:
         return errors
-    # Mark the value as being validated.
-    value.__schema_being_validated = True
+    VALIDATED_VALUES.__dict__[id(value)] = True
     # (If we have gotten here, we know that `value` provides an interface
     # other than zope.interface.Interface;
     # iow, we can rely on the fact that it is an instance
@@ -476,22 +594,27 @@ def _validate_fields(schema, value, errors=None):
             if not IMethod.providedBy(schema[name]):
                 try:
                     attribute = schema[name]
-                    if IField.providedBy(attribute):
+                    if IChoice.providedBy(attribute):
+                        # Choice must be bound before validation otherwise
+                        # IContextSourceBinder is not iterable in validation
+                        bound = attribute.bind(value)
+                        bound.validate(getattr(value, name))
+                    elif IField.providedBy(attribute):
                         # validate attributes that are fields
                         attribute.validate(getattr(value, name))
-                except ValidationError, error:
+                except ValidationError as error:
                     errors.append(error)
-                except AttributeError, error:
+                except AttributeError as error:
                     # property for the given name is not implemented
                     errors.append(SchemaNotFullyImplemented(error))
     finally:
-        delattr(value, '__schema_being_validated')
+        del VALIDATED_VALUES.__dict__[id(value)]
     return errors
 
 
+@implementer(IObject)
 class Object(Field):
     __doc__ = IObject.__doc__
-    implements(IObject)
 
     def __init__(self, schema, **kw):
         if not IInterface.providedBy(schema):
@@ -524,20 +647,19 @@ class Object(Field):
         super(Object, self).set(object, value)
 
 
+@implementer(IBeforeObjectAssignedEvent)
 class BeforeObjectAssignedEvent(object):
     """An object is going to be assigned to an attribute on another object."""
 
-    implements(IBeforeObjectAssignedEvent)
-
     def __init__(self, object, name, context):
         self.object = object
         self.name = name
         self.context = context
 
 
+@implementer(IDict)
 class Dict(MinMaxLen, Iterable):
     """A field representing a Dict."""
-    implements(IDict)
     _type = dict
     key_type = None
     value_type = None
@@ -577,222 +699,3 @@ class Dict(MinMaxLen, Iterable):
         if clone.value_type is not None:
             clone.value_type = clone.value_type.bind(object)
         return clone
-
-
-_isuri = re.compile(
-    # scheme
-    r"[a-zA-z0-9+.-]+:"
-    # non space (should be pickier)
-    r"\S*$").match
-
-
-class URI(BytesLine):
-    """URI schema field
-    """
-
-    implements(IURI, IFromUnicode)
-
-    def _validate(self, value):
-        """
-        >>> uri = URI(__name__='test')
-        >>> uri.validate("http://www.python.org/foo/bar")
-        >>> uri.validate("DAV:")
-        >>> uri.validate("www.python.org/foo/bar")
-        Traceback (most recent call last):
-        ...
-        InvalidURI: www.python.org/foo/bar
-        """
-
-        super(URI, self)._validate(value)
-        if _isuri(value):
-            return
-
-        raise InvalidURI(value)
-
-    def fromUnicode(self, value):
-        """
-        >>> uri = URI(__name__='test')
-        >>> uri.fromUnicode("http://www.python.org/foo/bar")
-        'http://www.python.org/foo/bar'
-        >>> uri.fromUnicode("          http://www.python.org/foo/bar")
-        'http://www.python.org/foo/bar'
-        >>> uri.fromUnicode("      \\n    http://www.python.org/foo/bar\\n")
-        'http://www.python.org/foo/bar'
-        >>> uri.fromUnicode("http://www.python.org/ foo/bar")
-        Traceback (most recent call last):
-        ...
-        InvalidURI: http://www.python.org/ foo/bar
-        """
-        v = str(value.strip())
-        self.validate(v)
-        return v
-
-
-_isdotted = re.compile(
-    r"([a-zA-Z][a-zA-z0-9_]*)"
-    r"([.][a-zA-Z][a-zA-z0-9_]*)*"
-    # use the whole line
-    r"$").match
-
-
-class Id(BytesLine):
-    """Id field
-
-    Values of id fields must be either uris or dotted names.
-    """
-
-    implements(IId, IFromUnicode)
-
-    def _validate(self, value):
-        """
-        >>> id = Id(__name__='test')
-        >>> id.validate("http://www.python.org/foo/bar")
-        >>> id.validate("zope.app.content")
-        >>> id.validate("zope.app.content/a")
-        Traceback (most recent call last):
-        ...
-        InvalidId: zope.app.content/a
-        >>> id.validate("http://zope.app.content x y")
-        Traceback (most recent call last):
-        ...
-        InvalidId: http://zope.app.content x y
-        """
-        super(Id, self)._validate(value)
-        if _isuri(value):
-            return
-        if _isdotted(value) and "." in value:
-            return
-
-        raise InvalidId(value)
-
-    def fromUnicode(self, value):
-        """
-        >>> id = Id(__name__='test')
-        >>> id.fromUnicode("http://www.python.org/foo/bar")
-        'http://www.python.org/foo/bar'
-        >>> id.fromUnicode(u" http://www.python.org/foo/bar ")
-        'http://www.python.org/foo/bar'
-        >>> id.fromUnicode("http://www.python.org/ foo/bar")
-        Traceback (most recent call last):
-        ...
-        InvalidId: http://www.python.org/ foo/bar
-        >>> id.fromUnicode("      \\n x.y.z \\n")
-        'x.y.z'
-
-        """
-        v = str(value.strip())
-        self.validate(v)
-        return v
-
-
-class DottedName(BytesLine):
-    """Dotted name field.
-
-    Values of DottedName fields must be Python-style dotted names.
-    """
-
-    implements(IDottedName)
-
-    def __init__(self, *args, **kw):
-        """
-        >>> DottedName(min_dots=-1)
-        Traceback (most recent call last):
-        ...
-        ValueError: min_dots cannot be less than zero
-
-        >>> DottedName(max_dots=-1)
-        Traceback (most recent call last):
-        ...
-        ValueError: max_dots cannot be less than min_dots
-
-        >>> DottedName(max_dots=1, min_dots=2)
-        Traceback (most recent call last):
-        ...
-        ValueError: max_dots cannot be less than min_dots
-
-        >>> dotted_name = DottedName(max_dots=1, min_dots=1)
-
-        >>> from zope.interface.verify import verifyObject
-        >>> verifyObject(IDottedName, dotted_name)
-        True
-
-        >>> dotted_name = DottedName(max_dots=1)
-        >>> dotted_name.min_dots
-        0
-
-        >>> dotted_name = DottedName(min_dots=1)
-        >>> dotted_name.max_dots
-        >>> dotted_name.min_dots
-        1
-        """
-        self.min_dots = int(kw.pop("min_dots", 0))
-        if self.min_dots < 0:
-            raise ValueError("min_dots cannot be less than zero")
-        self.max_dots = kw.pop("max_dots", None)
-        if self.max_dots is not None:
-            self.max_dots = int(self.max_dots)
-            if self.max_dots < self.min_dots:
-                raise ValueError("max_dots cannot be less than min_dots")
-        super(DottedName, self).__init__(*args, **kw)
-
-    def _validate(self, value):
-        """
-        >>> dotted_name = DottedName(__name__='test')
-        >>> dotted_name.validate("a.b.c")
-        >>> dotted_name.validate("a")
-        >>> dotted_name.validate("   a")
-        Traceback (most recent call last):
-        ...
-        InvalidDottedName:    a
-
-        >>> dotted_name = DottedName(__name__='test', min_dots=1)
-        >>> dotted_name.validate('a.b')
-        >>> dotted_name.validate('a.b.c.d')
-        >>> dotted_name.validate('a')
-        Traceback (most recent call last):
-        ...
-        InvalidDottedName: ('too few dots; 1 required', 'a')
-
-        >>> dotted_name = DottedName(__name__='test', max_dots=0)
-        >>> dotted_name.validate('a')
-        >>> dotted_name.validate('a.b')
-        Traceback (most recent call last):
-        ...
-        InvalidDottedName: ('too many dots; no more than 0 allowed', 'a.b')
-
-        >>> dotted_name = DottedName(__name__='test', max_dots=2)
-        >>> dotted_name.validate('a')
-        >>> dotted_name.validate('a.b')
-        >>> dotted_name.validate('a.b.c')
-        >>> dotted_name.validate('a.b.c.d')
-        Traceback (most recent call last):
-        ...
-        InvalidDottedName: ('too many dots; no more than 2 allowed', 'a.b.c.d')
-
-        >>> dotted_name = DottedName(__name__='test', max_dots=1, min_dots=1)
-        >>> dotted_name.validate('a.b')
-        >>> dotted_name.validate('a')
-        Traceback (most recent call last):
-        ...
-        InvalidDottedName: ('too few dots; 1 required', 'a')
-        >>> dotted_name.validate('a.b.c')
-        Traceback (most recent call last):
-        ...
-        InvalidDottedName: ('too many dots; no more than 1 allowed', 'a.b.c')
-
-        """
-        super(DottedName, self)._validate(value)
-        if not _isdotted(value):
-            raise InvalidDottedName(value)
-        dots = value.count(".")
-        if dots < self.min_dots:
-            raise InvalidDottedName("too few dots; %d required" % self.min_dots,
-                                    value)
-        if self.max_dots is not None and dots > self.max_dots:
-            raise InvalidDottedName("too many dots; no more than %d allowed" %
-                                    self.max_dots, value)
-
-    def fromUnicode(self, value):
-        v = str(value.strip())
-        self.validate(v)
-        return v
